@@ -21,7 +21,7 @@
  * SECTION:champlain-file-cache
  * @short_description: Stores and loads cached tiles from the file system
  *
- * #ChamplainFileCache is a map source that stores and retrieves tiles from the
+ * #ChamplainFileCache is a cache that stores and retrieves tiles from the
  * file system. Tiles most frequently loaded gain in "popularity". This popularity
  * is taken into account when purging the cache.
  */
@@ -374,12 +374,13 @@ champlain_file_cache_init (ChamplainFileCache *file_cache)
  * @cache_dir: the directory where the cache is created. For temporary caches
  * one more directory with random name is created inside this directory.
  * When cache_dir == NULL, a cache in ~/.cache/champlain is used.
+ * @renderer: the #ChamplainRenderer used for tiles rendering
  *
  * Constructor of #ChamplainFileCache.
  *
  * Returns: a constructed #ChamplainFileCache
  *
- * Since: 0.6
+ * Since: 0.8
  */
 ChamplainFileCache *
 champlain_file_cache_new_full (guint size_limit,
@@ -510,52 +511,41 @@ typedef struct
 {
   ChamplainMapSource *map_source;
   ChamplainTile *tile;
-  gchar *filename;
-  gulong handler;
-} TileLoadedCallbackData;
+} FileLoadedData;
 
 static void
-tile_loaded_cb (ClutterTexture *texture,
-    const GError *error,
-    TileLoadedCallbackData *user_data)
+tile_rendered_cb (ChamplainTile *tile,
+    ChamplainRenderCallbackData *data,
+    FileLoadedData *user_data)
 {
   ChamplainMapSource *map_source = user_data->map_source;
-  gchar *filename = user_data->filename;
-  ChamplainTile *tile = user_data->tile;
-  ChamplainFileCache *file_cache = CHAMPLAIN_FILE_CACHE (map_source);
-  ChamplainMapSource *next_source = champlain_map_source_get_next_source (map_source);
-  ChamplainFileCachePrivate *priv = file_cache->priv;
+  GFile *file;
+  ChamplainFileCache *file_cache;
+  ChamplainMapSource *next_source;
+  ChamplainFileCachePrivate *priv;
   GFileInfo *info = NULL;
-  GFile *file = NULL;
-  ClutterActor *actor = CLUTTER_ACTOR (texture);
   GTimeVal modified_time = { 0, };
+  gchar *filename = NULL;
 
-  g_signal_handler_disconnect (texture, user_data->handler);
+  g_signal_handlers_disconnect_by_func (tile, tile_rendered_cb, user_data);
+  g_slice_free (FileLoadedData, user_data);
 
-  if (tile)
-    g_object_remove_weak_pointer (G_OBJECT (tile), (gpointer *) &user_data->tile);
+  next_source = champlain_map_source_get_next_source (map_source);
+  file_cache = CHAMPLAIN_FILE_CACHE (map_source);
+  priv = file_cache->priv;
 
-  g_free (user_data);
-
-  if (!tile)
+  if (data->error)
     {
-      DEBUG ("Tile destroyed while loading");
-      if (!error && actor)
-        clutter_actor_destroy (actor);
-      goto cleanup;
-    }
-
-  if (error)
-    {
-      DEBUG ("Failed to load tile %s, error: %s", filename, error->message);
+      DEBUG ("Tile rendering failed");
       goto load_next;
     }
 
-  champlain_tile_set_content (tile, actor);
   champlain_tile_set_state (tile, CHAMPLAIN_STATE_LOADED);
 
-  /* Retrieve modification time */
+  filename = get_filename (file_cache, tile);
   file = g_file_new_for_path (filename);
+
+  /* Retrieve modification time */
   info = g_file_query_info (file,
       G_FILE_ATTRIBUTE_TIME_MODIFIED,
       G_FILE_QUERY_INFO_NONE, NULL, NULL);
@@ -570,7 +560,7 @@ tile_loaded_cb (ClutterTexture *texture,
 
   /* Notify other caches that the tile has been filled */
   if (CHAMPLAIN_IS_TILE_CACHE (next_source))
-    on_tile_filled (CHAMPLAIN_TILE_CACHE (next_source), tile);
+    champlain_tile_cache_on_tile_filled (CHAMPLAIN_TILE_CACHE (next_source), tile);
 
   if (tile_is_expired (file_cache, tile))
     {
@@ -611,6 +601,7 @@ tile_loaded_cb (ClutterTexture *texture,
   else
     {
       /* Tile loaded and no validation needed - done */
+      champlain_tile_set_fade_in (tile, FALSE);
       champlain_tile_set_state (tile, CHAMPLAIN_STATE_DONE);
       champlain_tile_display_content (tile);
       goto cleanup;
@@ -628,10 +619,52 @@ load_next:
 
 cleanup:
   g_free (filename);
+  g_object_unref (tile);
   g_object_unref (map_source);
 }
 
 
+static void
+file_loaded_cb (GFile *file,
+    GAsyncResult *res,
+    FileLoadedData *user_data)
+{
+  gboolean ok;
+  gchar *contents;
+  gsize length;
+  GError *error = NULL;
+  ChamplainTile *tile = user_data->tile;
+  ChamplainMapSource *map_source = user_data->map_source;
+  ChamplainRenderer *renderer;
+
+  ok = g_file_load_contents_finish (file, res, &contents, &length, NULL, &error);
+
+  if (!ok)
+    {
+      gchar *path;
+
+      path = g_file_get_path(file);
+      DEBUG ("Failed to load tile %s, error: %s", path, error->message);
+      g_free (path);
+      contents = NULL;
+      length = 0;
+      g_error_free (error);
+    }
+
+  g_object_unref (file);
+
+  renderer = champlain_map_source_get_renderer (map_source);
+
+  g_return_if_fail (CHAMPLAIN_IS_RENDERER (renderer));
+
+  g_signal_connect (tile, "render-complete", G_CALLBACK (tile_rendered_cb), user_data);
+
+  champlain_renderer_set_data (renderer, contents, length);
+  g_free (contents);
+  champlain_renderer_render (renderer, tile);
+}
+
+
 static void
 fill_tile (ChamplainMapSource *map_source,
     ChamplainTile *tile)
@@ -639,37 +672,39 @@ fill_tile (ChamplainMapSource *map_source,
   g_return_if_fail (CHAMPLAIN_IS_FILE_CACHE (map_source));
   g_return_if_fail (CHAMPLAIN_IS_TILE (tile));
 
+  ChamplainMapSource *next_source = champlain_map_source_get_next_source (map_source);
+
+  if (champlain_tile_get_state (tile) == CHAMPLAIN_STATE_DONE)
+    return;
+
   if (champlain_tile_get_state (tile) != CHAMPLAIN_STATE_LOADED)
     {
-      ChamplainFileCache *file_cache = CHAMPLAIN_FILE_CACHE (map_source);
-      TileLoadedCallbackData *callback_data;
-      ClutterTexture *texture;
+      FileLoadedData *user_data;
+      gchar *filename;
+      GFile *file;
 
-      callback_data = g_new (TileLoadedCallbackData, 1);
-      callback_data->tile = tile;
-      callback_data->map_source = map_source;
-      callback_data->filename = get_filename (file_cache, tile);
+      filename = get_filename (CHAMPLAIN_FILE_CACHE (map_source), tile);
+      file = g_file_new_for_path (filename);
+      g_free (filename);
 
-      DEBUG ("fill of %s", callback_data->filename);
+      user_data = g_slice_new (FileLoadedData);
+      user_data->tile = tile;
+      user_data->map_source = map_source;
 
-      g_object_add_weak_pointer (G_OBJECT (tile), (gpointer *) &callback_data->tile);
+      g_object_ref (tile);
       g_object_ref (map_source);
 
-      /* Load the cached version */
-      texture = CLUTTER_TEXTURE (clutter_texture_new ());
-      callback_data->handler = g_signal_connect (texture, "load-finished",
-          G_CALLBACK (tile_loaded_cb),
-          callback_data);
-      clutter_texture_set_load_async (texture, TRUE);
-      clutter_texture_set_from_file (texture, callback_data->filename, NULL);
-    }
-  else
-    {
-      ChamplainMapSource *next_source = champlain_map_source_get_next_source (map_source);
+      DEBUG ("fill of %s", filename);
 
-      /* Previous cache in the chain is validating its contents */
-      if (CHAMPLAIN_IS_MAP_SOURCE (next_source))
-        champlain_map_source_fill_tile (next_source, tile);
+      g_file_load_contents_async (file, NULL, (GAsyncReadyCallback) file_loaded_cb, user_data);
+    }
+  else if (CHAMPLAIN_IS_MAP_SOURCE (next_source))
+    champlain_map_source_fill_tile (next_source, tile);
+  else if (champlain_tile_get_state (tile) == CHAMPLAIN_STATE_LOADED)
+    {
+      /* if we have some content, use the tile even if it wasn't validated */
+      champlain_tile_set_state (tile, CHAMPLAIN_STATE_DONE);
+      champlain_tile_display_content (tile);
     }
 }
 
@@ -702,10 +737,11 @@ refresh_tile_time (ChamplainTileCache *tile_cache,
 
       g_file_info_set_modification_time (info, &now);
       g_file_set_attributes_from_info (file, info, G_FILE_QUERY_INFO_NONE, NULL, NULL);
-    }
+
+      g_object_unref (info);
+   }
 
   g_object_unref (file);
-  g_object_unref (info);
 
   if (CHAMPLAIN_IS_TILE_CACHE (next_source))
     champlain_tile_cache_refresh_tile_time (CHAMPLAIN_TILE_CACHE (next_source), tile);
@@ -829,6 +865,7 @@ on_tile_filled (ChamplainTileCache *tile_cache,
     }
 
 call_next:
+  g_free (filename);
   if (CHAMPLAIN_IS_TILE_CACHE (next_source))
     champlain_tile_cache_on_tile_filled (CHAMPLAIN_TILE_CACHE (next_source), tile);
 }
@@ -952,7 +989,7 @@ champlain_file_cache_purge (ChamplainFileCache *file_cache)
       const char *filename;
       guint size;
 
-      filename = sqlite3_column_text (stmt, 0);
+      filename = (const char *) sqlite3_column_text (stmt, 0);
       size = sqlite3_column_int (stmt, 1);
       highest_popularity = sqlite3_column_int (stmt, 2);
       DEBUG ("Deleting %s of size %d", filename, size);
